//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//            Bayesian Inference
// 
// Filename:  bayesapp.c
//
// Purpose:   Test operation of BayeSys with
//            UserEmpty,UserTry1,UserTry2,UserInsert1,UserInsert2,UserDelete1.
//=============================================================================
//
// BayeSys3 can operate with arbitrary likelihood functions.
// In this example, an atom has 2 coordinates interpreted as
//            location in [0,1] = Cube[0]  (restricted to [0.0.75])
// and
//         flux in [0,infinity) = -10.0*log(Cube[1])
// to correspond with testing the operation of MassInf with FluxUnit0=10
//
// These are only example programs, involving inefficient calculations that
// repeatedly build an object from scratch with "UserBuild" instructions.
//
//=============================================================================
// History:   JS       2 Jan 2002, 4 Jan 2003, 10 Feb 2003, 20 Aug 2003
//            Copyright (c) 2002,2003 Maximum Entropy Data Consultants Ltd.
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <stdlib.h>
#include <math.h>
#include "bayesys3.h"
#include "userstr.h"
#include "dimension.h"
#include "structure.h"
#include "fonction.h"
#include "constant.h"

double getLhood0();	// defined in o_chi.c

static int UserBuild(double*, CommonStr*, ObjectStr*, int, int);

//=============================================================================
//                           SIMPLE CODE
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserEmpty
//
// Purpose:   Set Lhood = logLikelihood(coords) for empty sample
//            with 0 atoms in Object, and initialise its other information.
//
//            Lhood := log L(empty)
//
//            I have already put the number 0 in Object->Natoms,
//            so you can use a simple "UserBuild" instruction.
//-----------------------------------------------------------------------------
// 
int UserEmpty(        //   O  >=0, or -ve error code
double*    Lhood,     //   O  loglikelihood
CommonStr* Common,    // I O  general information
ObjectStr* Object)    // I O  sample object, output new Lhood
{
    UserBuild(Lhood, Common, Object, 0, 1);
    return 1;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserTry1
//
// Purpose:   d(logLikelihood(coords)) after supposedly adding one new atom
//            to Object, without any side effects on it.
//
//            If Valency = 0,
//              dLtry := d logL(...,x)
//            else mimic the above with                                cool
//              dLtry := (1/cool) log INTEGRAL dPrior(z) dlogL(...,x,z)
//
//            I have already put the new atom x after the last location
//            of the Atoms list, at Object->Cubes[ Object->Natoms ], so
//            if Valency=0 you can use simple "UserBuild" instructions.
//-----------------------------------------------------------------------------
// 
int UserTry1(         //   O  +ve = OK, 0 = DO NOT USE, -ve = error
double*    dLtry,     //   O  trial d(logLikelihood) value
CommonStr* Common,    // I    general information
ObjectStr* Object)    // I    sample object (DO NOT UPDATE Lhood)
{
    int    Natoms = Object->Natoms;
    double Lhood  = Object->Lhood;                      // existing Lhood
    double Ltry;
    int    OK;

    *dLtry = 0.0;
    OK = UserBuild(&Ltry, Common, Object, Natoms+1, 0); // trial
    if( OK > 0 )
        *dLtry = Ltry - Lhood;                          // increment
    return  OK;                                         // OK?
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserTry2
//
// Purpose:   d(logLikelihood(coords)) after supposedly adding one,
//            then two new atoms to Object, without any side effects on it.
//
//   If Valency = 0,
//     dLtry1 := d logL(...,x1)
//
//     dLtry2 := d logL(...,x1,x2)
//
//   else mimic the above with                                    cool
//     dLtry1 := (1/cool) log INTEGRAL dPrior(z1) dlogL(...,x1,z1)
//                                                                         cool
//     dLtry2 := (1/cool) log INTEGRAL dPrior(z1,z2) dlogL(...,x1,z1,x2,z2)
//
//            I have already put the new atoms x1,x2 after the last location
//            of the Atoms list, at Object->Cubes[ Object->Natoms ] and
//                               at Object->Cubes[ Object->Natoms + 1 ]
//            so if Valency=0 you can use simple "UserBuild" instructions.
//-----------------------------------------------------------------------------
// 
int UserTry2(         //   O  +ve = OK, 0 = DO NOT USE, -ve = error
double*    dLtry1,    //   O  trial d(logLikelihood) value for 1st atom
double*    dLtry2,    //   O  trial d(logLikelihood) value for both atoms
CommonStr* Common,    // I    general information
ObjectStr* Object)    // I    sample object (DO NOT UPDATE Lhood)
{
    int    Natoms = Object->Natoms;
    double Lhood  = Object->Lhood;                           // existing Lhood
    double Ltry1, Ltry2;
    int    OK;

    *dLtry1 = *dLtry2 = 0.0;
    OK = UserBuild(&Ltry1, Common, Object, Natoms+1, 0);     //trial for 1 more
    if( OK > 0 )
    {
        *dLtry1 = Ltry1 - Lhood;                             // increment for 1
        OK = UserBuild(&Ltry2, Common, Object, Natoms+2, 0); //trial for 2 more
        if( OK > 0 )
            *dLtry2 = Ltry2 - Lhood;                         // increment for 2
    }
    return  OK;                             // return OK only if both trials OK
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserInsert1
//
// Purpose:   Insert 1 new atom into Object, keeping it up-to-date, and
//            set d(loglikelihood(coords)). 
//
//            If Valency = 0,
//              dL := d logL(...,x)
//
//            else
//                                                cool
//              sample z from  Prior(z) L(...,x,z)
//
//              and set    dL := d logL(...,x,z)  at the sampled z.
//
//            I have already put the new atom x at the last location of
//            the updated Atoms list, at Object->Cubes[ Object->Natoms - 1 ],
//-----------------------------------------------------------------------------
// 
int UserInsert1(      //   O  >=0, or -ve error code
double*    dL,        //   O  d(loglikelihood)
CommonStr* Common,    // I    general information
ObjectStr* Object)    // I O  sample object
{
    int    Natoms = Object->Natoms;              // new number
    double Lold   = Object->Lhood;               // not yet updated
    double Lnew;

    UserBuild(&Lnew, Common, Object, Natoms, 1); // new updated state
    *dL = Lnew - Lold;                           // I will update Object->Lhood
    return 1;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserInsert2
//
// Purpose:   Insert 2 new atoms into Object, keeping it up-to-date, and
//            set d(loglikelihood(fluxes)). 
//
//            If Valency = 0,
//              dL := d logL(...,x1,x2)
//
//            else
//                                                                cool
//              sample z1,z2 from  Prior(z1,z2) L(...,x1,z1,x2,z2)
//
//              and set  dL := d logL(...,x1,z1,x2,z2)  at the sampled z1,z2.
//
//            I have already put the new atoms x1,x2 at the last location
//            of the Atoms list, at Object->Cubes[ Object->Natoms - 2 ] and
//                               at Object->Cubes[ Object->Natoms - 1 ]
//            so if Valency=0 you can use a simple "UserBuild" instruction.
//-----------------------------------------------------------------------------
// 
int UserInsert2(      //   O  >=0, or -ve error code
double*    dL,        //   O  d(loglikelihood)
CommonStr* Common,    // I    general information
ObjectStr* Object)    // I O  sample object
{
    int    Natoms = Object->Natoms;              // new number
    double Lold   = Object->Lhood;               // not yet updated
    double Lnew;

    UserBuild(&Lnew, Common, Object, Natoms, 1); // new updated state
    *dL = Lnew - Lold;                           // I will update Object->Lhood
    return 1;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserDelete1
//
// Purpose:   Delete 1 old atom from Object, keeping it up-to-date, and
//            set d(loglikelihood(fluxes)). 
//  
//            dL := d logL(...)
//
//            I have already put the old atom after the last location of
//            the updated Atoms list, at Object->Cubes[ Object->Natoms ],
//            so if Valency=0 you can use a simple "UserBuild" instruction.
//-----------------------------------------------------------------------------
// 
int UserDelete1(      //   O  >=0, or -ve error code
double*    dL,        //   O  d(loglikelihood)
CommonStr* Common,    // I    general information
ObjectStr* Object)    // I O  sample object
{
    int    Natoms = Object->Natoms;              // new number
    double Lold   = Object->Lhood;               // not yet updated
    double Lnew;

    UserBuild(&Lnew, Common, Object, Natoms, 1); // new updated state
    *dL = Lnew - Lold;                           // I will update Object->Lhood
    return 1;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserBuild
//
// Purpose:   Build Lhood, and any other data info such as Mock from scratch.
//            Can be called with any number of atoms
//
//            This example implementation has
//            Mock[0] = SUM[atoms] flux ,             flux = - 10*log(Cube[1])
//            Mock[1] = SUM[atoms] flux * location ,  location = Cube[0]
//-----------------------------------------------------------------------------
// 
static int UserBuild( //   O  +ve = OK, 0 = DO NOT USE, -ve = error
double*    Lhood,     //   O  loglikelihood 
CommonStr* Common,    // I    General information
ObjectStr* Object,    // I(O) Cubes in, perhaps Mock out
int        Natoms,    // I    # atoms
int        output)    // I    +ve if Mock to be written out, -ve if not
{
	double** Cubes  = Object->Cubes;  // I    Cubes in [0,1)  [Natoms][Ndim]
	double*  Cube=Cubes[0];
	int valid;
	UserCommonStr* UserCommon = Common->UserCommon;
	
	valid = 0.; // DO NOT USE
	*Lhood = 0.;
	
	if( Natoms == 1 ) 
	{	
		/* Set the clumps parameters from the cube and eventually reject the whole object*/
		valid = rescaleCube_1Atom(Cube,Common->Ndim);
		
		/* Compute the likelyhood for this object */
		if( valid )
		{
			UserCommon->Nchi2++;
			*Lhood = o_chi();
			if( *Lhood == -1 )
				return 0;
			
			*Lhood =  -0.5 * (*Lhood);
			*Lhood += -0.5 * getLhood0();			
		}
	}
	return valid;
	
}

//=============================================================================
//       Dummy procedure to link to BayeSys3 when not running MassInf
//=============================================================================
int UserFoot(
double*    Cube,
CommonStr* Common,
int*       ibits,
double*    zbits,
int*       nbits)
{
    return (Cube && Common && ibits && zbits && nbits) ? 1 : 1;
}

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function: getParName
// 
// Purpose:  Return the name of the parameter corresponding to the ipx parameters
//=============================================================================
char* getParName(int ipx, char* name, int type)
{
	switch(ipx)
	{
		case(CX):
			strcpy( name, "x (arcsec)");
			break;
		case(CY):
			strcpy( name, "y (arcsec)");
			break;
		case(EPOT):
			if( type == 14 )
				strcpy( name, "gamma" );
			else
				strcpy( name, "emass" );
			break;
		case(THETA):
			strcpy( name, "theta (deg)" );
			break;
		case(RC):
			if( type == 13 )
				strcpy( name, "Re (arcsec)" );
			else if( type == 12 )
				strcpy( name, "rs (arcsec)" );
			else
				strcpy( name, "rc (arcsec)" );
			break;
		case(B0):
			if( type == 13 )
				strcpy( name, "Sigma_e (10^8 Msol/kpc^2)" );
			else
				strcpy( name, "sigma (km/s)" );
			break;
		case(ALPHA):
			if( type == 13 )
				strcpy( name, "n" );
			else
				strcpy( name, "alpha" );
			break;
		case(BETA):
			if( type == 12 )
				strcpy( name, "c" );
			else		
				strcpy( name, "beta" );
			break;																																										
		case(RCUT):
			if( type == 12 )
				strcpy( name, "r200 (arcsec)" );
			else
				strcpy( name, "rcut (arcsec)" );
			break;
		case(MASSE):
			strcpy( name, "M200 (Msol)" );
			break;
		case(PMASS):
			if( type == 12 )
				strcpy( name, "rhos (Msol/Mpc^3)" );				
			else
				strcpy( name, "pmass" );
			break;
		case(OMEGAM):
			strcpy( name, "omegaM" );
			break;
		case(OMEGAX):
			strcpy( name, "omegaX" );
			break;
		case(WX):
			strcpy( name, "wX" );
			break;			
	};
	return name;
}


/* Convert input Lenstool values to the corresponding values in <bayes.dat> file.
 * bayes2lt() function in readBayesModels.c
 */
static double lt2bayes(int i, int ipx, double val)
{
	extern struct pot	    lens[];
	extern struct g_cosmo C;

	switch(ipx)
	{
		case(THETA):
			val *= RTD;
			break;
	      case(B0):
	      	// copied from o_print_res()
			if ( lens[i].type <= 1 )
				val=sqrt(val/4./pia_c2);
			else if( lens[i].type == 13 )
				val=val*1e4*(cH0_4piG*C.h/distcosmo1(lens[i].z)); // in 10^8 Msol/kpc^2
			else
				val=sqrt(val/6./pia_c2);

	      	break;
/* Everything in solar masses
	      case(PMASS):
			if( lens[i].type == 12 )
				val /= 1e14; // rhos in 10^14 Msol -> Msol
			break;
		case(MASSE):
			if( lens[i].type == 12 )
				val /= 1e14; // masse in 10^14 Msol -> Msol
			break;
*/
		default:
			break;
	}
	
	return val;
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function: getParVal
// 
// Purpose:  Return the parameter value corresponding to the ilens and ipx parameters
//=============================================================================
static double getParVal(int i, int ipx)
{
	extern struct pot		lens[];
	double x;
	
	x = o_get_lens(i, ipx);
	
	return lt2bayes(i, ipx, x);
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Function:  UserMonitor
//
// Purpose:   I have provided a new ensemble of objects.
//
//        1.  For probabilistic exploration, restrict Common->cool to <= 1.
//                    cool = 0 is prior, seen at the first call.
//                0 < cool < 1 is "burn-in"
//                    cool = 1 is evolve posterior,
//                    cool -> infinity is maximum likelihood
//            Return with a POSITIVE return code to exit BayeSys,
//            usually after the end of an evolution stage with cool = 1.
//            You can use Common->Nsystem, which counts calls to UserMonitor.
//
//        2.  You should reset any nuisance parameters x in UserCommon and/or
//            each of your UserObject structures, preferably by sampling from
//            their conditonal probabilities Pr(x|Objects) and/or Pr(x|Object).
//
//        3.  Collect your statistics and display your diagnostics. 
//-----------------------------------------------------------------------------
//
int UserMonitor(            //   O  0 = continue, +ve = finish, -ve = abort
      CommonStr* Common,    // I    Full general information
      ObjectStr* Objects)   // I    Full ensemble of sample objects  [ENSEMBLE]
{
	extern struct g_mode	M;
	extern struct g_grille	G;
	extern struct g_image	I;
	extern struct g_pot		P;    
	extern int				block[][NPAMAX];
	extern int				cblock[NPAMAX];
	extern struct z_lim		zlim[];    
	extern struct galaxie	multi[NIMAX][NIMAX];
	extern int				optInterrupt; // Global variable modified in o_run_bayes.c/signalReset()
	
// control
	int     Nrem;    // # to go, as (2 * Nsample++) catches up with Nsystem++
    
	extern  struct  pot     lens[NLMAX];
    
    
// desired statistics
	UserCommonStr* UserCommon = (UserCommonStr*)Common->UserCommon;

// local
	ObjectStr*  Object;         // individual object
	int         k, i, j;           // counters
	int         ilens, ipx;
	double      chi2;
	FILE        *bayes;
	int         nimages;
	char        limages[ZMBOUND][IDSIZE];
	
//	 FILE    *debug;
//.............................................................................
// Re-set any user parameters in UserCommon or UserObject (none in this code).
// You will not need to reset the loglikelihoods.
/*
      debug = fopen("debug.dat", "a");
        for( ipx = 0; ipx < Common->ENSEMBLE; ipx++)
        {
                rescaleCube_1Atom(Objects[ipx].Cubes[0],Common->Ndim);
                fprintf( debug, "1 %lf %lf %lf %lf %lf\n", o_chi(), Common->cool,
                        sqrt(lens[0].b0/6./7.20997e-6), C.omegaX, C.wX);
        }
        fclose(debug);
*/


//.............................................................................
// USER IMPOSES FINAL TEMPERATURE AND TERMINATION CONDITION !
	if( Common->cool >= 1.0 )
		Common->cool = 1.0;
		
	rescaleCube_1Atom(Objects[0].Cubes[0], Common->Ndim);
	chi2  = o_chi();
//	if( getLhood0() != 0. ) chi2 = -0.5 * ( chi2 + getLhood0() );
	
	if( Common->cool < 1.0 )
	{
		Nrem = Common->Nsystem;
		
		fprintf(stdout,                     // (stderr flushes output immediately)
			"Burn-in : %10.6lf %5d %14.4lf %14.4lf\r", 
	    	Common->cool, Nrem, chi2, Common->Evidence);
	}
	else
	{	
		if( M.itmax == 0 )
			M.itmax = Common->Nsystem;
			
		Nrem = M.itmax - UserCommon->Nsample;  // or any other setting
		fprintf(stdout, "Sampling : %10.6lf %7d/%d %14.4lf %14.4lf[CTRL-C to interrupt]\r",  
			Common->cool, UserCommon->Nsample+1, M.itmax+1, chi2, Common->Evidence);
	}
	fflush(stdout);

// SAVE: save the chi2 value
	if( Common->cool  >= 1. )
	{
#ifdef DEBUG
		bayes= fopen( "bayes.dbg.dat" , "a" );
#else
		bayes= fopen( "bayes.dat" , "a" );
#endif

		// For each object
		for( j = 0; j < Common->ENSEMBLE; j++ )
		{
			// rescale the cube of object_j
			rescaleCube_1Atom(Objects[j].Cubes[0], Common->Ndim);
			chi2  = o_chi();
			if( getLhood0() != 0. ) chi2 = -0.5 * ( chi2 + getLhood0() );
				
			fprintf( bayes, "%d %lf ", UserCommon->Nsample+1, chi2); //%lf, Common->Evidence 
	
	// SAVE: save lens parameters 	
			for( ilens = 0; ilens < G.no_lens; ilens++ )
				for( ipx = CX; ipx <= PMASS; ipx++ )
					if( block[ilens][ipx] != 0 )
						fprintf(bayes, "%lf ", getParVal(ilens, ipx));				

	// SAVE: save cosmological parameters	
			for( ipx = OMEGAM; ipx <= WX; ipx++ )
				if( cblock[ipx] != 0 ) 
					fprintf(bayes, "%lf ", getParVal(0, ipx));
	
	// SAVE: save zmlimit parameters
			for( k = 0; k < I.nzlim; k++ )
				if( zlim[k].opt != 0 )
				{
					// look for the images families corresponding to the zmlimit to optimize
					nimages = splitzmlimit(zlim[k].n, limages);
					i = 0;
					while( indexCmp( multi[i][0].n, limages[0] ) ) i++;
						fprintf( bayes, "%lf ", multi[i][0].z );
				}
						
	// SAVE : save pot parameters
			if( P.ircut != 0 )
				fprintf( bayes, "%lf ", P.cut );
			if( P.isigma != 0 )
				fprintf( bayes, "%lf ", P.sigma );
				
	// SAVE: the nuisance parameters
			if( I.dsigposAs != -1. )
				fprintf( bayes, "%lf ", sqrt(I.sig2pos) );
			if( I.dsigell != -1. )
				fprintf( bayes, "%lf ", sqrt(I.sig2ell) );

/*				
	// SAVE : save the gradx and grady values of the multiple images
			for( k = 0; k < I.n_mult; k++ )
				for( ipx = 0; ipx < I.mult[k]; ipx++ )
					fprintf( bayes, "%lf %lf ", 
						multi[k][ipx].Grad.x, multi[k][ipx].Grad.y);	
	*/
			fprintf( bayes, "\n" );
		}
		fclose(bayes);
	}// end of save parameters if cool>=1

			
//.............................................................................
// Set any NUISANCE PARAMETERS you may have here
// Could over-ride system's choices of FluxUnit (both Common and Objects) here
// Could CALIBRATE data here (corrupting Evidence and Information)
//.............................................................................
// BURN-IN
    if( Common->cool < 1.0 )  // final temperature not reached..
        return 0;             // ...normal return, keep going
//.............................................................................
// EVOLUTION: Accumulate any desired statistics
    for( k = 0; k < Common->ENSEMBLE; k++ )
    {
        Object = &Objects[k];
  // # atoms
        UserCommon->atoms += Object->Natoms;
    }

  // # samples of full ensemble
    UserCommon->Nsample ++;

    if( Nrem > 0 && !optInterrupt)                                 // not finished...
        return 0;                                  // ..normal return
//.............................................................................
// EXIT WITH POSITIVE FINISH CODE WHEN EVOLUTION IS DEEMED COMPLETE
	NPRINTF(stderr, "\n" ); 
    return 1;
}

