#include<stdlib.h>
#include<float.h>
#include<string.h>
#include<math.h>
#include<fonction.h>
#include<constant.h>
#include<dimension.h>
#include<structure.h>
#include"lt.h"

/****************************************************************/
/*		nom:		readBayesModels	*/
/*		auteur:		Eric Jullo			*/
/*		date:		07/08/06   			*/
/*		place:		ESO, Chile			*
 *
 * Read the bayes.dat file and return an array[nParam][val].
 * where nParam = [chi2 -> Last param]
 * 
 * It also analyse the header of bayes.dat and set
 * the <G.no_lens>, <G.nlens>, <block>, <cblock> and <P.ircut>,
 * <P.isigma> global variables to their corresponding values.
 * 
 * Return:
 * - the bayes.dat array and nParam and nVal the size of the array.
 * - 0 for nVal and nParam if bayes.dat cannot be read.
 * 
 * Warning : 
 *  - Theta in <bayes.dat> is in degree but is converted to radian in <array>
 *  - sigma in <bayes.dat> is in km/s but is converted according to the lens type
 *    in <array>.
 *  
 * Then, the user can generate an error mass map even if he doesn't perform
 * an optimisation.  
 */

static double lhood_mode; // If 1, bayes.dat contains #ln(Lhood), otherwise 0 

static void analyseHeader( char *line );
static double bayes2lt(int ilens, int ipx, double val);
static void convertArray(int nVal, double **array );
static double method(int n, int param, double **array);


double	**readBayesModels(int *nParam, int *nVal)
{	
	extern struct g_grille	G;
	extern struct g_image	I;
	extern struct g_pot		P;
	
	extern int block[][NPAMAX];
	extern int cblock[NPAMAX];
	
	FILE	 *bayes;
	double **array;
	char   line[LINESIZE];
	int    nLines;
	char   *pch;	// token to split a line
	int    j, ilens, ipx;

	// Initialize ouput variables
	*nParam = 0;
	*nVal = 0;

	// Read bayes.dat
#ifdef DEBUG
	bayes = fopen( "bayes.dbg.dat", "r");
#else
	bayes = fopen( "bayes.dat" , "r" );
#endif

	if( bayes == NULL )
	{
		//fprintf(stderr, "WARNING: Reading bayes.dat. No such file found.\n");
		return 0;
	}
	
	//
	// REINITIALIZE ALL VARIABLES
	//
	
	// Get G.no_lens from bayes.dat
	// Reset G and block global variables
	G.no_lens = 0;
	for( ilens = 0; ilens < G.nlens; ilens++)
		for( ipx = CX; ipx <= PMASS; ipx ++)
			block[ilens][ipx] = 0;

	// Reset the cblock global variables
	for( ipx = OMEGAM; ipx <=WX; ipx++)
		cblock[ipx] = 0;
		
	// Reset the redshift I.nzlim
	I.nzlim = 0;
	
	// Reset the potfile 
	P.ircut = 0;
	P.isigma = 0;
	
	// Reset the nuisance
	I.dsigposAs = -1.;
	I.dsigell = -1.;

	// Set lhood_mode in #Chi2 in the bayes.dat file (default)
	lhood_mode = 0;

	// Read the header
	nLines = wc(bayes);
	fgets( line, LINESIZE, bayes); line[strlen(line)-1] = '\0';
	while( !feof(bayes) && !ferror(bayes) && line[0] == '#' )
	{ 
		(*nParam)++;
		analyseHeader( line );
		fgets( line, LINESIZE, bayes); line[strlen(line)-1] = '\0';
	}

	// Initialize array
	*nParam = (*nParam) - 1; //because Nsample is excluded from array
	array = alloc_square_double( *nParam , nLines - (*nParam) - 1);
		
	// Read the data
	while( !feof(bayes) && !ferror(bayes) )
	{

		// Skip the fist column and read the 2nd one
		pch = strtok( line, " " ); 
		pch = strtok (NULL, " ");
	
		// Fill array[][*nVal]
		j = 0;
		while( pch != NULL )
		{
			sscanf( pch, "%lf", &array[j++][*nVal] );
			pch = strtok (NULL, " ");
		}

		fgets(line, LINESIZE, bayes); line[strlen(line)-1] = '\0';
		(*nVal) ++;
	}			  
	
	// Catch errors
	if( ferror(bayes) || ( feof(bayes) && *nVal == 0 ) ) 
	{
//		fprintf( stderr, "WARNING : reading bayes.dat\n" );
		*nVal = *nParam = 0;
		return array;
	}
	fclose(bayes);
	
	// Convert <array> to values directly usable in lenstool
	convertArray(*nVal, array);
	
	return array;
}

static int imethod;		// use a specified method as parameter estimator
						// -1 : mean value
						// -2 : median value
						// -3 : mode value
						// -4 : best chi2 value
						
/* Set all the lenstool parameters to the values contained in the
 * bayes.dat file at line iVal.
 * If iVal < 0 then iVal specifies the method corresponding to the imethod index.
 */
void	setBayesModel( int iVal, int nParam, int nVal, double **array)
{
	extern struct g_grille	G;
	extern struct g_cosmo	C;
	extern struct g_image	I;
	extern struct g_pot	P;
	
	extern struct galaxie	multi[NIMAX][NIMAX];
	extern struct z_lim	zlim[];
	extern struct pot		lens[];
	extern int	block[][NPAMAX];
	extern int	cblock[NPAMAX];
	
	int     nimages;
	char    limages[ZMBOUND][IDSIZE];	
	double  tmp;
	int     ilens, ipx;
	int     iParam;
	int     i,j,k,l;
	
	iParam = IDPARAM1;
	imethod = iVal;
	
	for( ilens = 0; ilens < G.no_lens; ilens++ )
	{
		for( ipx = CX; ipx <= PMASS; ipx++ )
			if( block[ilens][ipx] != 0 )
				o_set_lens( ilens, ipx, method(nVal, iParam++, array) );
			
		// TODO in o_set_lens() assign directly to sigma rather than b0
		if( block[ilens][B0] != 0 )
			lens[ilens].sigma = lens[ilens].b0;
			
		set_dynamics(ilens);
	}
	
	// Set the cosmological parameters
	for( ipx = OMEGAM; ipx <= WX; ipx++ )
		if( cblock[ipx] != 0 )
			o_set_lens( 0, ipx,  method(nVal, iParam++, array) );
	
	if( C.kcourb == 0. )
		C.omegaX = 1 - C.omegaM;
	else
		C.kcourb = C.omegaM + C.omegaX - 1;		
	
	// Set the optimized redshifts to the corresponding images
	for( i = 0 ; i < I.nzlim; i++ )
	{
		tmp = method(nVal, iParam++, array);
		nimages = splitzmlimit(zlim[i].n, limages);
		for( j = 0; j < nimages; j++ )
		{
			k = 0;
			// loop over the images families
			while( indexCmp( multi[k][0].n, limages[j] ) && k < I.n_mult ) k++;
			// set the reshift of all the images of the family
			for( l = 0; l < I.mult[k]; l++ )
				multi[k][l].z = tmp;
		}
	}
					
	//rescale the pot parameters
	if( P.ircut != 0 )
		P.cut = method(nVal, iParam++, array);
	
	if( P.isigma != 0 )
		P.sigma = method(nVal, iParam++, array);
	
	o_scale_pot();

	//set the best nuisance
	if( I.dsigposAs != -1. )
	{
		I.sigposAs = method(nVal, iParam++, array);
		I.sig2pos = I.sigposAs*I.sigposAs;
	}
	if( I.dsigell != -1. )
	{
		I.sigell = method(nVal, iParam++, array);
		I.sig2ell = I.sigell*I.sigell;
	}	
	// set the DLS/DS ratio for all the images according to the new cosmological parameters
	for( i = 0 ; i < I.n_mult; i++ )
		for( j = 0; j < I.mult[i]; j++ )
			multi[i][j].dr=dratio(lens[0].z,multi[i][0].z);
		
}

/* Analyse a line of the bayes.dat header and eventually
 * In the case of reading a best.par file, reset the 
 * <G.no_lens>, <block>, <cblock>, <zlim> and <P.ircut>, <P.isigma> 
 * global variables as if they were optimized
 */
 static void analyseHeader( char *line )
 {
 	extern struct g_grille	G;
	extern struct g_image	I;
	extern struct g_pot     P;
 	
 	extern struct pot       lens[];
 	extern struct z_lim     zlim[];
 	extern int block[][NPAMAX];
 	extern int cblock[NPAMAX];
 	
 	int   i, ipx;
 	char  *pch, *pch2, name[50];
 	int	updated; 	// flag to prevent the case when the clump is already set as optimised
 
 	// skip the '#'
 	line = line + 1;
 	
 	if( !strcmp(line, "Nsample")  ||
 	    !strcmp(line, "Evidence") ||
 	    !strcmp(line, "Chi2") )
 		return;

	if( !strcmp(line, "ln(Lhood)") )
	{
		lhood_mode = 1;
		return;
	}


 	// Get the first word
 	pch = strtok( line, " " );
 	
 	// Skip the "O" if it exists in the <bayes.dat> file... <pch> in bayes.dat
 	if( pch[0] == 'O' )
 		pch++;
 		
	// Look for the right lens to change
	updated = 0;
 	for( i = 0; i < G.nlens && updated == 0 ; i++ )
	{
	 	// ... and in Lenstool... <pch2> in Lenstool
	 	pch2 = lens[i].n;
	 	if( pch2[0] == 'O' )
	 		pch2++;

 		if( !strcmp( pch, pch2 ) )
 		{
 			pch = strtok( NULL, " " );  // skip the ':'
 			pch = strtok( NULL, " " );
 			// Look for the right parameter to change
 			for( ipx = CX; ipx <= PMASS; ipx++ )
 			{
 				pch2 = strtok( getParName(ipx, name, lens[i].type), " ");	// remove the unit if any
 				if( !strcmp( pch, pch2) && block[i][ipx] == 0  )
 					block[i][ipx] = 1;
 			}	
 			// Prevent the case when the clump is already set as optimised
 			for( ipx = CX; ipx <= PMASS; ipx++ )
 				updated += block[i][ipx];
 			
 			// Increment G.no_lens only once per clump
 			if( updated == 1 )
 				G.no_lens ++;
 		}
	}
	
	// Look for cosmological parameters
	if( !strcmp(pch, "omegaM") )
		cblock[OMEGAM] = 1;
	if( !strcmp(pch, "omegaX") )
		cblock[OMEGAX] = 1;
	if( !strcmp(pch, "wX") )
		cblock[WX] = 1;

	// Look for the redshifts
	if( !strcmp(pch, "Redshift") )
	{
		pch = strtok( NULL, " " ); // pch contains 'of'
		pch += 3;	// pch points on the last part of line
		strcpy(zlim[I.nzlim].n, pch);
		// Append an image to optimize
		zlim[I.nzlim].opt = 1;
		I.nzlim++;		
	}
	
	// Look for the Potfile
	if( !strcmp(pch, "Pot") )
	{
		pch = strtok( NULL, " " ); // pch contains rcut or sigma
		if( !strcmp(pch, "rcut") )
			P.ircut  = 1;
		if( !strcmp(pch, "sigma") )
			P.isigma  = 1;
	}
	
	// Look for the nuisance
	if( !strcmp(pch, "SigposArsec") )
		I.dsigposAs = 1.;	// this is not the correct sigma but it doesn't matter
					// since the values are written in the bayes.dat file
					
	if( !strcmp(pch, "Sigell") )
		I.dsigell = 1.;
 }

/* Convert input <bayes.dat> values to the corresponding values readily usable
 * in lenstool.
 */
static double bayes2lt(int i, int ipx, double val)
{
	extern struct g_cosmo C;
	extern struct pot	    lens[];

	switch(ipx)
	{
		case(THETA):
			val *= DTR;
			break;
	      case(B0):
	      	if( lens[i].type == 13 )
	      		val *=1e8;
	      	break;
/* now it is in solar mass in bayes.dat also
		case(PMASS):
			if( lens[i].type == 12 )
				val *= 1e14; // rhos in 10^14 Msol -> Msol
			break;
		case(MASSE):
			if( lens[i].type == 12 )
				val *= 1e14; // masse in 10^14 Msol -> Msol
			break;
*/
		default:
			break;
	}
	
	return val;
}

static void convertArray(int nVal, double **array ) 
{
 	extern struct g_grille   G;
 	extern struct pot        lens[];
 	extern int block[][NPAMAX];
 	
 	int   iVal, i, ipx;
 	int   nParam;
 	
 	for( iVal = 0; iVal < nVal; iVal++ )
 	{
 		nParam = IDPARAM1; // position of the first parameter in array
 					 // 1st column in array is chi2
 		if( lhood_mode ) 	array[0][iVal] *= -1.; // so that the best lhood has the lowest value
 							 // ie same behavior as chi2
 							 
		for( i = 0; i < G.no_lens; i++ )
			for( ipx = CX; ipx <= PMASS; ipx++ )
				if( block[i][ipx] != 0 )
				{
					array[nParam][iVal] = bayes2lt( i, ipx, array[nParam][iVal] );
					nParam++;
				}
 	}	
}

/* Return the value in <list> that has produced the best chi2*/
static double bestchi2(int n, double *chi2, double *list)
{
	double chi2min;
	int i, imin;
	
	imin = 0;
	chi2min = DBL_MAX;
	
	for( i = 0; i < n; i++)
		if( chi2[i] < chi2min )
		{
			chi2min = chi2[i];
			imin = i; 
		}
		
	return( list[imin] );
}

/* Return an estimated value of the values in <array> depending
 * on the value of the <imethod> value.
 * <imethod> may also be a line number in array[param]. 
 * n is the number of elements in array[param].
 */
static double method(int n, int param, double **array)
{
	double val;
	
	switch( imethod )
	{
		case(-1):
			val = mean(n, array[param]);
			break;			
		case(-2):
			val = median(n, array[param]);
			break;
		case(-3):
			val = mode(n, array[param]);
			break;
		case(-4):
			val = bestchi2(n, array[0], array[param]);
			break;
		default:
			val = array[param][imethod];
			break;
	}
	
	return val;
}
